/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   mo_fedotov@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>

#include <qdatetime.h>
#include <qtimer.h>

#include "logger.h"
#include "bt_comm.h"
#include "ar_client.h"
#include "ar_cmds.h"

extern logger2*  	theLog;
extern AR_Client* 	theClient;
extern AR_Parser*	conf;
extern KAnyRemote*  mainWin;

AR_Cmds::AR_Cmds() 
{
	timers 		= NULL;
	remoteOn	= true;
}

AR_Cmds::~AR_Cmds() 
{
	// free timers ??
	freeTimers();
}

void AR_Cmds::setStyle(bool isServer)
{
	isServerMode = isServer;
}

void AR_Cmds::setRemoteOn(bool set)
{
    theLog->info(QString("Switch remote ") + (set ? "On" : "Off"));
    remoteOn = set;
}

bool AR_Cmds::isRemoteOn(void)
{
    return remoteOn;
}

int AR_Cmds::execInFork(const char *cmd)
{
    int ret;
    int cpid = fork();
    if( cpid < 0) {
    	theLog->error("Can not fork!");
    	return -1;
    }

    if(cpid) {	    // father
    	waitpid(cpid,&ret,0);
    	if(ret!=-1 && WIFEXITED(ret)) {
	    ret = WEXITSTATUS(ret);
    	}
    } else {	    // child
    	theLog->info("Close port in child");
    	theClient->getComm().closePort(0);  	
    	ret = system(cmd);
    	if(ret!=-1 && WIFEXITED(ret)) {
	    ret = WEXITSTATUS(ret);
	}
    	exit(ret);
    }

    return ret;
}

int AR_Cmds::execSimpleCmd(QString command, int toFile, cmdParams *params)
{
    if (command.isNull() || command.isEmpty()) { // Empty command. Skip it.
        theLog->info("execSimpleCmd null input");
    	return EXIT_OK;
    }
    theLog->info("execSimpleCmd >" + command + "<");
    
 
    if (!isRemoteOn()) {
    	theLog->debug("Remote if OFF. Skip event");
    	return EXIT_OK;
    }
    
    int papameterized = 0;
    QString cmd("");
    
    while(!(command.isNull() || command.isEmpty())) {

        papameterized = command.find(PARAM_START);   // $(
 	
	if (papameterized < 0) {   	    
    	    theLog->debug("non-parametrized command (1)");
    	    cmd = cmd + command;
            break;
	} else if (command.mid(papameterized+2).startsWith(PARAM_CALLID)) {   	    // Need to substitute it with callerID
	    
            theLog->debug("parametrized command - callerID");
	    
            // First part
	    cmd = cmd + command.left(papameterized) + " ";
            
	    // Add params and second part
	    cmd = cmd + theClient->getCallerID();
            
            command = command.right(command.length() - papameterized - QString(PARAM_CALLID).length() - 2);
	
        } else if (command.mid(papameterized+2).startsWith(PARAM_TIME)) {     // Need to substitute it with time
	    
            theLog->debug("parametrized command - Time");

	    // First part
	    cmd = cmd + command.left(papameterized) + " ";
	    	
	    QString time = QTime::currentTime().toString().left(5);
		
	    // Add params and second part
	    cmd = cmd + time;
            
            command = command.right(command.length() - papameterized - QString(PARAM_TIME).length() - 2);
    	
        } else if (command.mid(papameterized+2).startsWith(PARAM_MODE)) {   // Need to substitute it with mode
    	
            theLog->debug("parametrized command - Mode");
	    
	    // First part
    	    cmd = cmd + command.left(papameterized);
	    
    	    // Add params and second part
    	    cmd = cmd + conf->getMode();
            
            command = command.right(command.length() - papameterized - QString(PARAM_MODE).length() - 2);	

    	} else if (command.mid(papameterized+2).startsWith(PARAM_PARAM)) {
            // This is dirty trick to make Bemused commands like VOLM<volume> works
	    // also it is used in Java Client list handling
           
            if (params == NULL || params->value.isNull() || params->value.isEmpty()) {
                theLog->error("Parameter value was not set!");
                return EXIT_OK;
            }
	    
	    // First part
	    cmd = cmd + command.left(papameterized);
	    
 	    // Add params
	    cmd = cmd + params->value;
	    
    	    // Add second part
    	    command = command.right(command.length() - papameterized - QString(PARAM_PARAM).length() - 2);	
            
     	} else if (command.mid(papameterized+2).startsWith(PARAM_INDEX)) {
            // Used in Java Client list handling

            if (params == NULL || params->index.isNull() || params->index.isEmpty()) {
                theLog->error("Parameter value was not set!");
                return EXIT_OK;
            }
	    
	    // First part
	    cmd = cmd + command.left(papameterized);
	    
 	    // Add params
	    cmd = cmd + params->index;
	    
    	    // Add second part
    	    command = command.right(command.length() - papameterized - QString(PARAM_INDEX).length() - 2);	
            
   	} else {    // Just copy
	    
    	    theLog->debug("non-parametrized command");
    	    cmd = cmd + command;
            break;
	}
    }
	
    if (toFile) {       // We needs to wait results, so do not run in background
        cmd += QString(" > ") + TOFILE;
    } else {
        if (!cmd.contains("&")) {
            cmd += QString(" &");
        }
    }
    theLog->debug("final cmd >" + cmd + "<");

    return (execInFork(cmd.ascii()) == -1 ? EXIT_NOK : EXIT_OK);
}

int AR_Cmds::processSendCmd(QString origCmd)
{
    int ret = EXIT_OK;
    
    QString tag  = origCmd.section(",",0,0).simplifyWhiteSpace();
    QString cmd  = origCmd.section(",",1);  // no simplifyWhiteSpace()
    
    if (tag.isNull() || tag.isEmpty() || cmd.isNull() || cmd.isEmpty()) {
        theLog->debug("processSendCmd(): incorrect input (1)");                
        return EXIT_NOK;
    }
    
    theLog->debug("processSendCmd() is >" + tag + "< >" + cmd + "<");
    
    if (tag == SEND_STRING) {
        ret = theClient->getComm().writeSocketConn(cmd);
        if (ret != EXIT_OK) {
	    theLog->debug("Fails in writeSocketConn()");
        }
    } else if (tag == SEND_BYTES) {
        ret = theClient->getComm().writeBytesSocketConn(cmd);
        if (ret != EXIT_OK) {
	    theLog->debug("Fails in writeBytesSocketConn()");
        }    
    } else {
        theLog->debug("processSendCmd(): incorrect input (3)");                
        return EXIT_NOK;
    }
    return ret;
}

void AR_Cmds::sendData(int type, QString tag, QString data) 
{
	int ret;
	if (type == ID_SEND && strcmp(tag,SEND_BYTES) == 0) {
		ret = theClient->getComm().writeBytesSocketConn(data);
	} else {
		ret = theClient->getComm().writeSocketConn(data);
	}
	if (ret != EXIT_OK) {
		theLog->debug("Fails in write(Bytes)SocketConn()");
	}
}

int AR_Cmds::execCmdAndGetResults(int type, QString commandOrig, timerCmd* timer, cmdParams *params) 
{
	theLog->debug("execCmdAndGetResults");

	if (isServerMode && isRemoteOn() && commandOrig != QString::null) {
       
		QString tag  = commandOrig.section(",",0,0).simplifyWhiteSpace();
		QString part1("");
		QString part2("");
		int cmdIdx = 1;
        
		/*
		if (tag.contains(SET_TEXT)) {
			part1 = commandOrig.section(",",1,1).simplifyWhiteSpace();
			cmdIdx = 2;
		} else 
		*/
		if (tag.contains(SET_TEXT) || tag.contains(SET_LIST) || tag.contains(SET_ILIST)) {
			part1 = commandOrig.section(",",1,1).simplifyWhiteSpace();
			part2 = commandOrig.section(",",2,2).simplifyWhiteSpace();
			cmdIdx = 3;
		}
        
		QString cmd  = commandOrig.section(",",cmdIdx); // No simplify white space
        
		if (tag.isNull() || tag.isEmpty() || cmd.isNull() || cmd.isEmpty()) {
			theLog->debug("execCmdAndGetResults(): incorrect input");                
			return EXIT_NOK;
		}
		if (type != ID_SET && type != ID_SEND ) {
			theLog->debug("execCmdAndGetResults(): incorrect input (2)");                
			return EXIT_NOK;
		}
        
		int ret = execSimpleCmd(cmd,1,params);
		if (ret != EXIT_OK) {
			theLog->debug("Return code from execSimpleCmd() is " + QString().setNum(ret));
			return ret;
		}    
        
		QString res = readResultsFile();
		theLog->debug("readResultsFile() " + res);

		if (timer != NULL && timer->lastValue != QString::null && timer->lastValue == res) {
			theLog->debug("No needs to send the same data.");
			return EXIT_OK;
        	}
        	if (timer != NULL) {    // Update "last data"
			updateTimerLastData(timer, res);
        	}

		if (type == ID_SEND) {  // Just send string
			sendData(type, tag, res);
		} else {  		// ID_SET
			QString chunk = QString(CMD_SET) + "(" + tag + ","; 	// For example, Set(title,

			if (strcmp(tag,SET_TEXT) == 0) {
				// long commands
				// Because of bug in Java String class in SE and Nokia (fails to get correct indexOf() 
				// if string longer than 256 characters) we have to separate result to several chunks
                
				int doLoop  = 1;   
				int headLen = chunk.length();     
				while (doLoop > 0) {
                
					if (!part1.isNull()) {
						if (part1 == QString("replace")) {
							if (doLoop == 1) {
								chunk += part1;
							} else {
								chunk += "add";
							}
						} else {
							chunk += part1;
						}    
					}
					if (!part2.isNull()) {
						chunk += "," + part2;
					}
					chunk += ",";
                    
					int curLen = chunk.length();
                    
					if (res.length() + curLen > 253) {    		// 253 + ");\0" = 256
						chunk += res.left(253 - curLen);
						res = res.mid(253 - curLen + 1);           	// and move pointer
						doLoop++;
					} else {                             		// last portion
						chunk += res;
						doLoop = -1;
					}
                
					chunk += ");";
            
					sendData(type, tag, chunk);
					chunk = chunk.left(headLen);
				}

			} else if (strcmp(tag,SET_LIST) == 0 || strcmp(tag,SET_ILIST) == 0) {
				// long commands
				// Because of bug in Java String class in SE and Nokia (fails to get correct indexOf() 
				// if string longer than 256 characters) we have to separate result to several chunks

				int doLoop = 1;        
				int headLen = chunk.length();     
				int section = 0;
				
				QString onePart = res.section(",",section,section);
				while (doLoop > 0) {
                
					if (!part1.isNull()) {
						if (part1 == QString("replace")) {
							if (doLoop == 1) {
								chunk += part1;
							} else {
								chunk += "add";
							}
						} else {
							chunk += part1;
						}    
					}
					if (!part2.isNull()) {
						chunk += "," + part2;
					}
					while ((!onePart.isNull()) && chunk.length() + onePart.length() < 252) {
						chunk += "," + onePart;
						section++;
						onePart = res.section(",",section,section);
					}
					if (onePart.isNull()) {
						doLoop = -1;    // Last portion
					} else {
						doLoop++;
					}

					chunk += ");";
            
					sendData(type, tag, chunk);
					chunk =  chunk.left(headLen);
				}
			} else {    // short commands. Are all of them really short (<256 bytes?)
				chunk += res + ");";
				sendData(type, tag, chunk);
			}
		}
	}
	return EXIT_OK;
}

int AR_Cmds::execCmd(int cmdType, QString commandIn, cmdParams *params)
{
    int ret;
    QString commandOrig;

    if (commandIn.isNull() || commandIn.isEmpty()) {
        theLog->info("execCmd >no params<");
        commandOrig = QString::null;
    } else {
        theLog->info("execCmd >" + commandIn + "<");
        commandOrig = commandIn;
    }    
    
    switch (cmdType) {
    
        case ID_EXIT:               
    	    if (isRemoteOn()) {	    // User asks to exit	
    	        theLog->info("Exit key detected");
    	        theClient->Disconnect();			    

    	        return EXIT_ABORT;
            }
            break;
            
        case ID_TMPDISCONN:
            if (isRemoteOn()) {            // User asks to temporary disconnect
    	        theLog->info("Temporary disconnect key detected");
				theClient->Disconnect();
                
    	        return EXIT_DISCON; // But what with the rest of the command line ? (now just skip)
            }
            break;
            
        case ID_REMOTEON:
            theLog->info("Switch remote on");
    	    setRemoteOn(true);
            break;
            
        case ID_REMOTEOFF:
            theLog->info("Switch remote off");
    	    setRemoteOn(false);
            break;
            
        case ID_EXEC:
            theLog->info("Command: Exec");
            if (isRemoteOn()) {
                execSimpleCmd(commandOrig, 0, params);
            }
            break;
            
        case ID_SETMODE:
            theLog->info("Command: SetMode");
            
	    if (isRemoteOn()) {           
	    	int i = 0;    
	    	type_key* k = conf->findItem(QString(EVT_EXITMODE), i, NULL);
    	    	if (k != NULL && i == FLAG_EXACT) {
    	            theLog->debug("Exec cmd on exit from mode");
    	            ret = handleCmdByKey(k, NULL);
    	            if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    	return ret;
    	            }
    	    	} else {
    	            theLog->debug("Leave mode\n");
    	    	}

                conf->setMode(commandOrig);

		i = 0; 
	    	k = conf->findItem(QString(EVT_ENTRMODE), i, NULL);
    	    	if (k != NULL && i == FLAG_EXACT) {
    	            theLog->debug("Exec cmd on enter to mode");
    	            ret = handleCmdByKey(k, NULL);
    	            if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    	return ret;
    	            }
    	    	} else {
    	            theLog->debug("Entered to mode");
    	    	}
            }
            break;

        case ID_SENDCKPD:
            theLog->info("Command: SendCKPD");
            if (!isServerMode && isRemoteOn()) {

                ret = theClient->getComm().sendSeq(commandOrig);
                if (ret != EXIT_OK) {
	            theLog->debug("Fails in sendSeq()");
                }
    
                if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	            return ret;
                }
            }    
            break;
            
        case ID_SEND:
            theLog->info("Command: Send");
            if (isServerMode && isRemoteOn() && commandOrig != QString::null) {
                ret = processSendCmd(commandOrig);
                if (ret != EXIT_OK) {
	            theLog->debug("Fails in processSendCmd()");
                }
            }    
            break;
            
        case ID_SET:          
            theLog->info("Command: Set");
            if (isServerMode && isRemoteOn() && commandOrig != QString::null) {
	    	QString tmptext(CMD_SET);
		tmptext += QString("(");
		
		int i = 0;
		while (true) {
			QString part  = commandOrig.section("\\n",i,i);
			if (part.isNull()) {
				break;
			}
			tmptext += part + "\n";
			i++;
		}
                tmptext += ");";
                //tmptext += QString("(") + commandOrig + ");";
               
                 ret = theClient->getComm().writeSocketConn(tmptext);
                if (ret != EXIT_OK) {
	            theLog->debug("Fails in writeSocketConn() (Set)");
                }
            }    
            break;
            
        case ID_EXECSET:   
            theLog->info("Command: ExecAndSet");
            
            execCmdAndGetResults(ID_SET, commandOrig, NULL, params);
            break;
            
        case ID_EXECSEND:   
            theLog->info("Command: ExecAndSend");
            
            execCmdAndGetResults(ID_SEND, commandOrig,NULL, NULL);
            break;

        case ID_TIMERCMD:   
            theLog->info("Command: Timer");
            if (isServerMode && isRemoteOn() && commandOrig != QString::null) {
                return processTimerCmd(commandOrig);
            }     
            break;

        case ID_CANCELTIMER:   
            theLog->info("Command: CancelTimer");
            if (isServerMode && isRemoteOn() && commandOrig != QString::null) {
                cancelTimer(commandOrig);
            }  
            break;
            
        default:  
            theLog->info("Unknown command");                 
            break;
   }
   
   return EXIT_OK;
}

QString AR_Cmds::readResultsFile() 
{
	FILE *fp;
	struct stat buf;
	
	if(stat(TOFILE, &buf) == -1) {
		return NULL;
	}
	
	long fLen = buf.st_size;
	theLog->info("readResultsFile >" + QString().setNum(fLen) + "<");
	
	fp=fopen(TOFILE,"r");
	if (fp == NULL) {
		return NULL;
	}
    
    	char* fBuffer = (char*) calloc(fLen+1,1);
    
	fread(fBuffer, sizeof(char), fLen, fp);
    
    	QString b(fBuffer);
	free(fBuffer);
	
	return b;
}

int AR_Cmds::processCommands(cmdItem* cmdList, cmdParams *params) 
{
    while (cmdList) {
            
        int ret = execCmd(cmdList->type, cmdList->descr, params);
        if (ret != EXIT_OK) {
            theLog->debug("Return code from execCmd() is " + QString().setNum(ret));
 	     
            if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
  	        theLog->debug("Clean auto repeat flag after execCmd()");

                theClient->cleanupRepeatKeys();
		
	        return ret;
            }
        }
        cmdList = cmdList->next;
    }

    return EXIT_OK;
}

//
// Top-level command processing function
//

int AR_Cmds::handleCmdByKey(type_key *tk, cmdParams *params)
{
    if(tk == NULL) {
    	theLog->debug("handleCmdByKey() NULL input");
    	return EXIT_OK;
    }
    theLog->debug("handleCmdByKey() >" + tk->key + "<");
    
    cmdItem* cmdList = conf->getCommand(tk);
    if (cmdList == NULL) {
        theLog->debug("No appropriate command was found by getCommand()");
    } else {
     	theLog->debug("Ready to execute command");
     
        // In AT mode: User had pressed some key ... 
        // Send ToMainMenu sequence of CKPD's to show main menu again
        if (!isServerMode) {
            if (theClient->getComm().sendSeq(conf->getToMainMenu()) != EXIT_OK) {
    	        theLog->debug("Fails in sendSeq()");
            }
        }
        
        return processCommands(cmdList, params);
    }	
	
    return EXIT_OK;	
}
//
// Functions related to timer commands
//
int AR_Cmds::processTimerCmd(QString origCmd)
{
    int ret = EXIT_OK;
    
    theLog->debug("processTimerCmd(): >" + origCmd + "<");
    
    QString all = origCmd;
    
    QString id  = origCmd.section(",",0,0).simplifyWhiteSpace();
    QString t1  = origCmd.section(",",1,1).simplifyWhiteSpace();
    QString t2  = origCmd.section(",",2,2).simplifyWhiteSpace();
    QString cmd = origCmd.section(",",3);  // no simplifyWhiteSpace()

    if (id.isNull() || cmd.isNull() || t1.isNull() || t2.isNull()) {
        theLog->debug("processTimerCmd(): incorrect input");                
        return EXIT_NOK;
    }
    // Do not allow timer commands for Send(string/bytes...)
    if (!(cmd.startsWith(SET_ICONS)  ||
          cmd.startsWith(SET_TITLE)  ||
          cmd.startsWith(SET_STATUS) ||
          cmd.startsWith(SET_TEXT)   ||
          cmd.startsWith(SET_BG)     ||
          cmd.startsWith(SET_FG)     ||
          cmd.startsWith(SET_VIBRATE)||
          cmd.startsWith(SET_FONT)   ||
          cmd.startsWith(SET_LIST)   ||
          cmd.startsWith(SET_ILIST))) {
        theLog->debug("processTimerCmd(): incorrect tag in command");                
        return EXIT_NOK;
    }

    float tmout = t1.toFloat();
    int   times = t2.toInt();

    if (createTimer(id, ID_SET, tmout, times, cmd) == EXIT_NOK) {
        theLog->debug("processTimerCmd(): timer already exists or input was incorrect");                
        return EXIT_NOK;
    }
    
    timerCmd* tt = NULL;
    execCmdAndGetResults(ID_SET, cmd, findTimer(id, &tt), NULL);
	
    // AR_Cmds is not successor of QObject so send timeout to mainWin
    theLog->debug("Set commands timer");
    QTimer::singleShot(500, mainWin, SLOT(handleTimerCmd()));

    return ret;
}

timerCmd* AR_Cmds::findTimer(QString id, timerCmd **prev) 
{
    timerCmd* ptr = timers;
    
    while (ptr != NULL && (!ptr->id.startsWith(id))) {
        *prev = ptr;
        theLog->debug("findTimer(): set prev to >" + (*prev)->id + "<");
        ptr  = ptr->next;
    }

    return ptr;
}

int AR_Cmds::createTimer(QString id, int type, float tmout, int times, QString cmd)
{
    if (id == NULL || cmd == NULL) {
        theLog->debug("createTimer(): null input");                
        return EXIT_NOK;
    }
    
    timerCmd* ptr = NULL;
    if (findTimer(id, (timerCmd**) &ptr) != NULL) {
        theLog->debug("createTimer(): timer already exists");                
        return EXIT_NOK;
    } 
    timerCmd *tm = new timerCmd();
    
    tm->id        = id;
    tm->type      = type;
    tm->ticks     = 0;
    tm->times     = 1;
    tm->tmout     = tmout;
    tm->maxTimes  = times;
    tm->cmd       = cmd;
    tm->lastValue = QString::null;
    
    // Insert in head
    if (timers == NULL) { // Insert first
        tm->next = NULL;
    } else {
        tm->next = timers;
    }
    timers = tm;
    
    return EXIT_OK;
}

int AR_Cmds::cancelTimer(QString id) 
{
    theLog->debug("cancelTimer() >" + id + "<");
    
    timerCmd *prev = NULL;
    timerCmd *tm = findTimer(id, (timerCmd **) &prev);
    if (tm == NULL) {
        theLog->debug("Cant find timer to cancel!");
        return EXIT_NOK;
    }    
    if (tm == timers) { //remove in head
        timers = tm->next;
    } else {
        if (prev != NULL) {
            prev->next = tm->next;
        } else {
            theLog->error("Previous item absent for non-first element");
            return EXIT_NOK;
        }
    }        
    delete tm;
    
    return EXIT_OK;
}

void AR_Cmds::freeTimers()
{
    theLog->debug("freeTimers()");
    timerCmd *tm = timers;

    while (tm != NULL) {
        timerCmd *tmd = tm;
        tm =   tm->next;
        delete tmd;
    }
	timers = NULL;
    return;
}

int AR_Cmds::verifyTimers(int ticksInSec) 
{
    timerCmd *tm   = timers;
    
    int numTimers = 0;
    while (tm != NULL) {
        
        if (tm->ticks >= (tm->tmout * ticksInSec)) {
            theLog->debug("verifyTimers(): it is time to execute >" + tm->cmd + "<");

            execCmdAndGetResults(tm->type, tm->cmd, tm, NULL);
            
            tm->ticks = 0;
            
            if (tm->maxTimes > 0) {
                tm->times++;
                if (tm->times >= tm->maxTimes) {    // Cancel this timer 
                    theLog->debug("verifyTimers(): timer " + tm->id + " self-canceled");
                    
                    // Cancel this timer
                    timerCmd *tmc = tm;
                    
                    tm = tm->next;

                    cancelTimer(tmc->id);
                    
                    continue;
                }
            }
        } else {
            tm->ticks++;
        }
        //prev = tm;
        tm   = tm->next;
        numTimers++;
    }
    return numTimers;
}

void AR_Cmds::updateTimerLastData (timerCmd* timer, QString data)
{
	timer->lastValue = data;
}
